from __future__ import annotations

import escnn.group
from escnn.group import IrreducibleRepresentation
from escnn.group import directsum, change_basis
from escnn.group.irrep import restrict_irrep

from .so3_utils import *

from .utils import *

import numpy as np

from typing import Tuple, Callable, Iterable, List, Dict, Any, Union

try:
    import py3nj
except ImportError:
    import warnings
    warnings.warn("`py3nj` package not found! Will use a numerical method to compute the SO(3) Clebsh-Gordan coefficents. This is much slower but the coefficients will be cached on disk.")

    py3nj = None


__all__ = ["SO3"]


class SO3(Group):
    
    PARAM = PARAMETRIZATION

    PARAMETRIZATIONS = PARAMETRIZATIONS

    def __init__(self, maximum_frequency: int = 2):
        r"""
        Build an instance of the special orthogonal group :math:`SO(3)` which contains continuous rotations in the space.
        
        
        Subgroup Structure:
        
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |    ``id[0]``                      |    ``id[1]``                      |    subgroup                                                                                                                                                                                                   |
        +===================================+===================================+===============================================================================================================================================================================================================+
        |        'so3'                      |                                   |   :math:`SO(3)` group itself                                                                                                                                                                                  |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |        'ico'                      |                                   |   Icosahedral :math:`I` subgroup                                                                                                                                                                              |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |        'octa'                     |                                   |   Octahedral :math:`O` subgroup                                                                                                                                                                               |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |        'tetra'                    |                                   |   Tetrahedral :math:`T` subgroup                                                                                                                                                                              |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |        False                      |    -1                             |   :math:`SO(2)` subgroup of planar rotations around Z axis                                                                                                                                                    |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |        False                      |     N                             |   :math:`C_N` of N discrete planar rotations around Z axis                                                                                                                                                    |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |        True (or float)            |    -1                             |   *dihedral* :math:`O(2)` subgroup of planar rotations around Z axis and out-of-plane :math:`\pi` rotation around axis in the XY plane defined by ``id[1] / 2`` (X axis by default)                           |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |        True (or float)            |     N>1                           |   *dihedral* :math:`D_N` subgroup of N discrete planar rotations around Z axis and out-of-plane :math:`\pi` rotation around axis in the XY plane defined by ``id[1] / 2`` (X axis by default)                 |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |        True (or float)            |     1                             |   equivalent to ``(False, 2, adj)``                                                                                                                                                                           |
        +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        
        Args:
            maximum_frequency (int, optional): the maximum frequency to consider when building the irreps of the group
        
        """
        
        assert (isinstance(maximum_frequency, int) and maximum_frequency >= 0)
        
        super(SO3, self).__init__("SO(3)", True, False)
        
        self._maximum_frequency = maximum_frequency
        
        self._identity = self.element(IDENTITY, PARAMETRIZATION)
        
        self._build_representations()

    @property
    def generators(self) -> List[GroupElement]:
        raise ValueError(f'{self.name} is a continuous groups and '
                         f'some of its generators are infinitesimal. '
                         f'This is not currently supported')
    
    @property
    def identity(self) -> GroupElement:
        return self._identity

    @property
    def elements(self) -> List[GroupElement]:
        return None
     
    @property
    def _keys(self) -> Dict[str, Any]:
        return dict()
    
    @property
    def subgroup_trivial_id(self):
        return (False, 1)

    @property
    def subgroup_self_id(self):
        return 'so3'

    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################
    
    def _inverse(self, element, param=PARAM):
        r"""
        Return the inverse element of the input element
        """
        
        return _invert(element, param=param)

    def _combine(self, e1, e2,
                 param=PARAM,
                 param1=None,
                 param2=None
                 ):
        r"""
        Return the sum of the two input elements
        """
        return _combine(e1, e2, param=param, param1=param1, param2=param2)

    def _equal(self, e1, e2,
               param=PARAM,
               param1=None,
               param2=None,
               ) -> bool:
        r"""
        Check if the two input values corresponds to the same element
        """
        return _equal(e1, e2, param=param, param1=param1, param2=param2)

    def _hash_element(self, element, param: str = PARAM):
        return _hash(element, param)

    def _repr_element(self, element, param: str = PARAM) -> str:
        return _repr(element, param)

    def _is_element(self, element, param: str = PARAM, verbose: bool = False) -> bool:
    
        if not _check_param(element, param):
            if verbose:
                print(f"Element {element} is not a rotation")
            return False
    
        return True
    
    def _change_param(self, element, p_from: str, p_to: str):
        assert p_from in self.PARAMETRIZATIONS
        assert p_to in self.PARAMETRIZATIONS
        return _change_param(element, p_from, p_to)

    ###########################################################################
    
    def sample(self, param: str = PARAM) -> GroupElement:
        return self.element(_grid('rand', N=1, parametrization=param)[0])
    
    def grid(self, type: str, *args, adj: GroupElement = None, parametrization: str = PARAMETRIZATION, **kwargs) -> List[GroupElement]:
        r"""
        Method which builds different collections of elements of :math:`\SO3`.

        Depending on the value of ``type``, the method accept a different set of parameters using the ``args`` and the
        ``kwargs`` arguments:
        
        * ``type = "rand"``. Generate ``N`` *random* uniformly distributed samples over the group. The method accepts an integer ``N`` and, optionally, a random seed ``seed`` (or an instance of :class:`numpy.random.RandomState`).
        
        * ``type = "thomson"``.  Generate ``N`` samples distributed (approximately) uniformly over the group. The samples are obtained with a numerical method which tries to minimize a potential energy depending on the relative distance between the points. The first call of this method for a particular value of ``N`` might be slow due to this numerical method. However, the resulting set of points is cached on disk such that following calls will return the same result instantly.

        * ``type = "thomson_cube"``.  Generate ``24*N`` samples with *cubic* (*octahedral*) symmetry, distributed (approximately) uniformly over the group. The samples are obtained with a numerical method which tries to minimize a potential energy depending on the relative distance between the points. The first call of this method for a particular value of ``N`` might be slow due to this numerical method. However, the resulting set of points is cached on disk such that following calls will return the same result instantly.

        * ``type = "ico", "cube", "tetra"``.  Generate respectively ``60``, ``24`` or ``12`` samples corresponding to the rotational symmetries of respectively the Icosahedron (or Dodecahedron), Octahedron (or Cube) or the Tetrahedron.

        * ``type = "hopf"``.  Generates *about* ``N`` points by creating a HEAPlix grid on the sphere and combining it with a regular grid over :math:`\SO2` using Hopf fibration.
        
        .. todo ::
            add reference to paper explaining Hopf fibration based grid
                              
        * ``type = "fibonacci"``.  Generates *about* ``N`` points by creating a Fibonacci grid on the sphere and combining it with a regular grid over :math:`\SO2` using Hopf fibration.
        
        Args:
            type (str): string identifying the type of samples
            *args: arguments specific for the type of samples chosen
            adj (GroupElement, optional): optionally, apply an adjoint transform to the sampled elements.
            parametrization (str, optional):
            **kwargs: arguments specific for the type of samples chosen

        Returns:
            a list of group elements

        """
        if adj is None:
            adj = self.identity
        adj = adj.to(PARAMETRIZATION)

        return [
            self.element(g, param=parametrization)
            for g in _grid(type, *args, adj=adj, parametrization=parametrization, **kwargs)
        ]

    def sphere_grid(self, type: str, *args, adj: GroupElement = None, **kwargs) -> List[GroupElement]:
        r"""
        
        Method which builds different collections of points over the sphere.
        
        Here, a sphere is interpreted as the quotient space :math:`\SO3 / \SO2`.
        The list returned by this method contains instances of :class:`~escnn.group.GroupElement`.
        These are elements of :class:`~escnn.group.SO3` and should be interpreted as *representatives* of cosets in
        :math:`\SO3 / \SO2`.

        Depending on the value of ``type``, the method accept a different set of parameters using the ``args`` and the
        ``kwargs`` arguments:
        
        * ``type = "rand"``. Generate ``N`` *random* uniformly distributed samples over the sphere. The method accepts an integer ``N`` and, optionally, a random seed ``seed`` (or an instance of :class:`numpy.random.RandomState`).
        
        * ``type = "thomson"``.  Generate ``N`` samples distributed (approximately) uniformly over the sphere. The samples are obtained with a numerical method which tries to minimize a potential energy depending on the relative distance between the points. The first call of this method for a particular value of ``N`` might be slow due to this numerical method. However, the resulting set of points is cached on disk such that following calls will return the same result instantly.

        * ``type = "thomson_cube"``.  Generate ``24*N`` samples with perfect *cubic* (*octahedral*) symmetry, distributed (approximately) uniformly over the shere. The samples are obtained with a numerical method which tries to minimize a potential energy depending on the relative distance between the points. The first call of this method for a particular value of ``N`` might be slow due to this numerical method. However, the resulting set of points is cached on disk such that following calls will return the same result instantly.

        * ``type = "ico", "cube", "tetra"``.  Return the orbit of the point :math:`(0, 0, 1)` under the action of the Icosahedral, Octahedral or Tetrahedral group. The resulting samples correspond to the corners, edges or faces of a Platonic solid.
        
        .. todo ::
            add more types to support vertices, edges and faces for all solids.
            Currently, ``"ico"`` returns the centers of the 30 edges while ``"cube"`` returns the centers of its 6 faces

        .. warning::
            The naming convention might change in future to support faces/edges/vertices of the solids

        * ``type = "healpix"``.  Generates *about* ``N`` points using the HEAPlix grid on the sphere.
        
        * ``type = "fibonacci"``.  Generates ``N`` points using the Fibonacci grid on the sphere.
        
        * ``type = "longlat"``.  Generates an ``N`` by ``M`` grid over the sphere in the longitude (``N``) - latitude (``M``) coordinates.
        

        Args:
            type (str): string identifying the type of samples
            *args: arguments specific for the type of samples chosen
            adj (GroupElement, optional): optionally, apply an adjoint transform to the sampled elements.
            **kwargs: arguments specific for the type of samples chosen

        Returns:
            a list of group elements

        """
    
        if adj is None:
            adj = self.identity
        adj = adj.to('MAT')

        points = _sphere_grid(type, *args, adj=adj, **kwargs)

        elements = np.zeros((len(points), 3, 3))
        elements[:, :, 2] = points
        y = np.random.randn(len(points), 3) - points
        elements[:, :, 1] = y - (y*points).sum(axis=1, keepdims=True)*points
        elements[:, :, 1] /= np.linalg.norm(elements[:, :, 1], axis=1, keepdims=True)
        elements[:, :, 0] = np.cross(elements[:, :, 2], elements[:, :, 1])
        det = np.linalg.det(elements)
        assert np.allclose(np.abs(det), 1.)
        elements[:, :, 2] *= det.reshape(-1, 1)

        return [
            self.element(g, param='MAT') for g in elements
        ]

    def testing_elements(self, n=3) -> Iterable[GroupElement]:
        r"""
        A finite number of group elements to use for testing.
        """
        return self.grid('ico')
        
        samples = np.empty((n**3, 3))
        i = 0
        for a in range(n):
            alpha = 2 * np.pi * a / n
            for b in range(n):
                # beta = np.pi * (2 * b + 1) / (2*n)
                # beta = np.pi * (b + 1) / (n+2)
                # sample around b = 0  and b = np.pi to check singularities
                beta = np.pi * b / n
                for c in range(n):
                    gamma = 2 * np.pi * c / n
                    samples[i, :] = alpha, beta, gamma
                    i+=1

        samples = self._change_param(samples, p_from='ZYZ', p_to=self.PARAM)

        # samples = set()
        # i = 0
        # for a in range(n):
        #     alpha = 2 * np.pi * a / n
        #     for b in range(n):
        #         # beta = np.pi * (2 * b + 1) / (2*n)
        #         # beta = np.pi * (b + 1) / (n+2)
        #         # sample around b = 0  and b = np.pi to check singularities
        #         beta = np.pi * b / n
        #         for c in range(n):
        #             gamma = 2 * np.pi * c / n
        #             i+=1
        #             sample = alpha, beta, gamma
        #
        #             sample = self._change_param(sample, p_from='ZYZ', p_to='Q')
        #
        #             sign = np.heaviside(sample, 1.)*2 -1.
        #             sample = sample * sign.prod()
        #
        #             norm = np.linalg.norm(sample)
        #             assert np.allclose(norm, 1.), (alpha, beta, gamma, sample, sign, norm)
        #
        #             samples.add(tuple(sample.tolist()))
        #
        # samples = np.array(list(samples))
        # samples = self._change_param(samples, p_from='Q', p_to=PARAMETRIZATION)
        
        samples = [self.element(g, self.PARAM) for g in samples]
        
        return samples

    def __eq__(self, other):
        if not isinstance(other, SO3):
            return False
        else:
            return self.name == other.name # and self._maximum_frequency == other._maximum_frequency

    def _process_subgroup_id(self, id):

        if not isinstance(id, tuple):
            id = (id,)
    
        if not isinstance(id[-1], GroupElement):
            id = (*id, self.identity)
    
        assert id[-1].group == self

        if len(id) == 3 and isinstance(id[0], float):
            # O(2) (or subgroup) subgroup
            # if id[0] is bool, it specifies SO(2) vs O(2) subgroup
            # if id[0] is float, it assumes O(2) and it is twice the flip axis of O(2)
            # in this last case, we convert it to the boolean convention and include the flip axis inside the adjoing

            # The factor of 2 in id[0] comes from the fact that a flip around the axis theta, as an element of O(2), is
            # the combination of a reflection along the X axis (theta=0) and a rotation by 2*theta.
            # The values id[0] = 2*theta represents the SO(2) component of the flip.

            flip_axis = id[0] / 2.
            flip = np.asarray([0., 0., np.sin(flip_axis / 2), np.cos(flip_axis / 2)])
            flip = self.element(flip, 'Q')
            adjoint = (~flip) @ id[-1]
            id = (True, id[1], adjoint)
        
        assert isinstance(id[0], bool) or isinstance(id[0], str), id[0]

        if len(id) == 3 and id[0] == True and id[1] == 1:
            # flip subgroup of the O(2) subgroup of SO(3)
            # this is equivalent to the C_2 subgroup of 180 deg rotations out of the plane
            change_axis = np.asarray([0., np.sin(-np.pi / 4), 0., np.cos(-np.pi / 4)])
            adj = self.element(change_axis, 'Q') @ id[-1]
            id = (False, 2, adj)
    
        return id

    def _subgroup(self, id: Tuple) -> Tuple[
        Group,
        Callable[[GroupElement], GroupElement],
        Callable[[GroupElement], GroupElement]
    ]:
        r"""
        id can either be
            - a string in ['ico', 'octa' or 'tetra']
        or
            - a tuple [flip (out of plane 180 rot), N (number of discrete rots)]
        
        id can also be a tuple, whose first elements contain the id above, and the last is an element of SO(3) used
        to perform adjunction
        
        
        Returns:
            a tuple containing

                - the subgroup,

                - a function which maps an element of the subgroup to its inclusion in the original group and

                - a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)
              
        """
        
        sg_id, adj = id[:-1], id[-1]
        assert isinstance(adj, GroupElement) and adj.group == self
        
        if len(sg_id) == 1 and isinstance(sg_id[0], str):
            sg_id = sg_id[0]
            if sg_id == 'so3':
                sg = self
                # parent_mapping = lambda x, adj=adj, _adj=~adj: _adj @ x @ adj
                parent_mapping = build_adjoint_map(self, ~adj)
                # child_mapping = lambda x, adj=adj, _adj=~adj: adj @ x @ _adj
                child_mapping = build_adjoint_map(self, adj)
            elif sg_id == 'ico':
                sg = escnn.group.ico_group()
                parent_mapping = rotgroup_to_so3(adj, sg)
                child_mapping = so3_to_rotgroup(adj, sg)
            elif sg_id == 'octa':
                sg = escnn.group.octa_group()
                parent_mapping = rotgroup_to_so3(adj, sg)
                child_mapping = so3_to_rotgroup(adj, sg)
            elif sg_id == 'tetra':
                # sg = Tetrahedral()
                # parent_mapping = ico_to_so3(adj, sg)
                # child_mapping = so3_to_ico(adj, sg)
                raise NotImplementedError()
            else:
                raise ValueError(f'Subgroup "{sg_id}" not recognized!')
            
        elif len(sg_id) == 2:
            # planar subgroups
            
            if sg_id == (False, -1):
                # SO(2)
                sg = escnn.group.so2_group(self._maximum_frequency)
                parent_mapping = so2_to_so3(adj, sg)
                child_mapping = so3_to_so2(adj, sg)
            elif sg_id == (True, -1):
                # O(2)

                sg = escnn.group.o2_group(self._maximum_frequency)
                parent_mapping = o2_to_so3(adj, sg)
                child_mapping = so3_to_o2(adj, sg)
            elif sg_id[0] == False and sg_id[1] > 0:
                # Cyclic subgroup
                # or flip subgroup of the O(2) subgroup
                n = sg_id[1]
                assert isinstance(n, int)
                sg = escnn.group.cyclic_group(n)
                parent_mapping = so2_to_so3(adj, sg)
                child_mapping = so3_to_so2(adj, sg)

            elif sg_id[0] == True and sg_id[1] > 1:
                # Dihedral subgroup
                n = sg_id[1]
                assert isinstance(n, int)
                sg = escnn.group.dihedral_group(n)

                parent_mapping = o2_to_so3(adj, sg)
                child_mapping = so3_to_o2(adj, sg)

            elif sg_id[0] == True and sg_id[1] == 1:
                # flip subgroup of the O(2) subgroup
                # This case should have already been caught in the CyclicGroup case (thanks to the
                # process_subgroup_id method)
                raise ValueError(f'Subgroup "{sg_id}": this case should have already been caught!')
            else:
                raise ValueError(f'Subgroup "{sg_id}" not recognized!')
        else:
            raise ValueError(f'Subgroup "{sg_id}" not recognized!')

        return sg, parent_mapping, child_mapping

    def _combine_subgroups(self, sg_id1, sg_id2):
    
        sg_id1 = self._process_subgroup_id(sg_id1)
        sg1, inclusion, restriction = self.subgroup(sg_id1)
        sg_id2 = sg1._process_subgroup_id(sg_id2)
    
        sg_id1, adjoint1 = sg_id1[:-1], sg_id1[-1]
    
        adjoint2 = None
        sg_id = None
    
        if sg_id1 == ('so3',):
            # subgroup of SO3
            sg_id = sg_id2[:-1]
            adjoint2 = inclusion(sg_id2[-1])

        elif sg_id1[0] == 'ico' and len(sg_id1) == 1:
            # I
            if sg_id2[0] in ['ico', 'tetra']:
                sg_id = sg_id2[0],
                adjoint2 = inclusion(sg_id2[1])
            else:
                flip, n, adjoint_ico = sg_id2
                sg_id = flip, n

                adjoint2 = inclusion(adjoint_ico)

                if n == 3:
                    axis = np.asarray([1., 1., 1.]) / np.sqrt(3)
                    rot = find_rotation_pole2point(axis)

                    adjoint2 = ~self.element(rot, 'MAT') @ adjoint2
                elif n == 5:
                    from escnn.group.groups.ico import _PHI
                    axis = np.asarray([_PHI, 0., 1.])
                    axis /= np.linalg.norm(axis)
                    rot = find_rotation_pole2point(axis)
                    adjoint2 = ~self.element(rot, 'MAT') @ adjoint2
                else:
                    assert n in [1, 2]

        elif sg_id1[0] in ['octa', 'tetra'] and len(sg_id1) == 1:
            # O
            # T
            raise NotImplementedError()

        # planar subgroups
        elif sg_id1[0] == False and isinstance(sg_id1[1], int):
            # SO(2) or C_N
            sg_id = False, sg_id2

        elif sg_id1[0] == True and isinstance(sg_id1[1], int):
            # O(2) or D_N
            flip = sg_id2[0] is not None
            if flip and sg_id1[1] == -1:
                sg_id = sg_id2
            elif flip and sg_id1[1] > 0:
                sg_id = sg_id2[0] * 2 * np.pi / sg_id1[1], sg_id2[1]
            else:
                sg_id = False, sg_id2[1]

        elif sg_id1[0] is True and sg_id1[1] == 1:
            # flip subgroup of O(2), C_2
            # This case should have already been caught in the CyclicGroup case (thanks to the
            # process_subgroup_id method)
            # this is equivalent to C_2 above
            raise ValueError(f'Subgroup "{sg_id1}": this case should have already been caught!')
        else:
            raise ValueError(f'Subgroup "{sg_id1}" not recognized!')
    
        if adjoint2 is not None:
            adjoint = adjoint1 @ adjoint2
        else:
            adjoint = adjoint1
    
        return sg_id + (adjoint,)

    def _restrict_irrep(self, irrep: Tuple, id) -> Tuple[np.matrix, List[Tuple]]:
        r"""
        
        Returns:
            a pair containing the change of basis and the list of irreps of the subgroup which appear in the restricted irrep
            
        """
        
        sg_id, adj = id[:-1], id[-1]
        
        irr = self.irrep(*irrep)
        l = irr.attributes['frequency']

        sg, _, _ = self.subgroup(id)

        irreps = []
        change_of_basis = None
        
        try:
            if len(sg_id) == 1 and isinstance(sg_id[0], str):
                sg_id = sg_id[0]
                if sg_id == 'so3':
                    change_of_basis = irr.change_of_basis
                    irreps = irr.irreps
                elif sg_id == 'ico':
                    if l <= 2:
                        change_of_basis = np.eye(irr.size)
                        irreps = [(l,)]
                    else:
                        raise NotImplementedError()
                elif sg_id == 'octa':
                    if l <= 1:
                        change_of_basis = np.eye(irr.size)
                        irreps = [(l,)]
                    else:
                        raise NotImplementedError()
                elif sg_id == 'tetra':
                    raise NotImplementedError()
                else:
                    raise ValueError(f'Subgroup "{sg_id}" not recognized!')

            elif len(sg_id) == 2:
                # planar subgroups
        
                if sg_id == (False, -1):
                    # SO(2)
                    irreps = [(f,) for f in range(l+1)]
                    change_of_basis = np.zeros((2*l+1, 2*l+1))
                    change_of_basis[l, 0] = 1.
                    for f in range(l):
                        change_of_basis[l+f+1, 2*f+1] = 1.
                        change_of_basis[l-f-1, 2*f+2] = 1.

                elif sg_id == (True, -1):
                    # O(2)
                    j = l % 2
                    irreps = [(j, 0)] + [(1, f) for f in range(1, l+1)]
                    change_of_basis = np.zeros((2*l+1, 2*l+1))
                    change_of_basis[l, 0] = 1.
                    for f in range(l):
                        change_of_basis[l + f + 1, 2 * f + 1] = 1.
                        change_of_basis[l - f - 1, 2 * f + 2] = 1.
                        
                        if (l+f) % 2 == 0:
                            # TODO - prove this!
                            block = change_of_basis[[l - f - 1, l+f+1], 2*f+1:2*f+3]
                            block = block @ np.asarray([
                                [0., -1],
                                [1., 0.]
                            ])
                            change_of_basis[[l - f - 1, l + f + 1], 2 * f + 1:2 * f + 3] = block
                            
                elif sg_id[0] == False and sg_id[1] > 0:
                    # Cyclic Group
                    
                    # restrict first to SO(2) and then reuse the irreps restriction from SO(2) to C_N
                    
                    so2_sgid = self._process_subgroup_id((False, -1))
                    change_of_basis_so2, irreps_so2 = self._restrict_irrep(irrep, so2_sgid)
                    so2, _, _ = self.subgroup(so2_sgid)
                    
                    p = 0
                    cob_so2_cn = np.zeros((2*l+1, 2*l+1))
                    for irr in irreps_so2:
                        cob_cn, irr_cn = so2._restrict_irrep(irr, sg_id[1])
                        irreps += irr_cn
                        
                        irr = so2.irrep(*irr)
                        cob_so2_cn[p:p+irr.size, p:p+irr.size] = cob_cn
                        p += irr.size
                    
                    change_of_basis = change_of_basis_so2 @ cob_so2_cn

                elif sg_id[0] == True and sg_id[1] > 1:
                    # Dihedral Group
                    
                    # restrict first to O(2) and then reuse the irreps restriction from O(2) to D_N

                    o2_sgid = self._process_subgroup_id((True, -1))
                    change_of_basis_o2, irreps_o2 = self._restrict_irrep(irrep, o2_sgid)
                    o2, _, _ = self.subgroup(o2_sgid)

                    p = 0
                    cob_o2_cn = np.zeros((2 * l + 1, 2 * l + 1))
                    for irr in irreps_o2:
                        cob_cn, irr_cn = o2._restrict_irrep(irr, (0., sg_id[1]))
                        irreps += irr_cn
        
                        irr = o2.irrep(*irr)
                        cob_o2_cn[p:p + irr.size, p:p + irr.size] = cob_cn
                        p += irr.size

                    change_of_basis = change_of_basis_o2 @ cob_o2_cn
                    
                elif sg_id[0] == True and sg_id[1] == 1:
                    # flip subgroup of the O(2) subgroup
                    # This case should have already been caught in the CyclicGroup case (thanks to the
                    # process_subgroup_id method)
                    raise ValueError(f'Subgroup "{sg_id}": this case should have already been caught!')
                else:
                    raise ValueError(f'Subgroup "{sg_id}" not recognized!')
            else:
                raise ValueError(f'Subgroup "{sg_id}" not recognized!')
            
        except NotImplementedError:
            change_of_basis, irreps = restrict_irrep(irr, sg_id)

        change_of_basis = self.irrep(*irrep)(adj).T @ change_of_basis

        return change_of_basis, irreps

    def _build_representations(self):
        r"""
        Build the irreps for this group

        """
    
        # Build all the Irreducible Representations
    
        k = 0
    
        # add Trivial representation
        self.irrep(k)
    
        # add other irreducible representations
        for k in range(1, self._maximum_frequency + 1):
            self.irrep(k)
    
        # add all the irreps to the set of representations already built for this group
        self.representations.update(**{irr.name: irr for irr in self.irreps()})

    @property
    def trivial_representation(self) -> Representation:
        return self.irrep(0)

    def bl_regular_representation(self, L: int) -> Representation:
        r"""
        Band-Limited regular representation up to frequency ``L`` (included).

        Args:
            L(int): max frequency

        """
        name = f'regular_{L}'
        
        if name not in self._representations:
            irreps = []
            
            for l in range(L+1):
                irreps += [self.irrep(l)]*(2*l+1)
            
            self._representations[name] = directsum(irreps, name=name)

        return self._representations[name]

    def bl_quotient_representation(self,
                                   L: int,
                                   subgroup_id,
                                   ) -> escnn.group.Representation:
        r"""
        Band-Limited quotient representation up to frequency ``L`` (included).
        
        The quotient representation corresponds to the action of the current group :math:`G` on functions over the
        homogeneous space :math:`G/H`, where :math:`H` is the subgroup of :math:`G` identified by ``subgroup_id``.
        
        Args:
            L(int): max frequency
            subgroup_id: id identifying the subgroup H.

        """

        name = f"quotient[{subgroup_id}]_{L}"

        if name not in self.representations:
            subgroup, _, _ = self.subgroup(subgroup_id)

            homspace = self.homspace(subgroup_id)

            irreps = []

            for l in range(L+1):
                irr = self.irrep(l)
                multiplicity = homspace.dimension_basis(irr.id, homspace.H.trivial_representation.id)[1]
                irreps += [irr] * multiplicity

            self.representations[name] = directsum(irreps, name=name)

        return self.representations[name]

    def bl_sphere_representation(self,
                                 L: int,
                                 ) -> escnn.group.Representation:
        r"""
        Representation of :math:`\SO3` acting on functions on the sphere band-limited representation up to frequency
        ``L`` (included).

        Args:
            L(int): max frequency
        """
        return self.bl_quotient_representation(L, (False, -1))

    def bl_induced_representation(self,
                                  L: int,
                                  repr: escnn.group.IrreducibleRepresentation,
                                  subgroup_id,
                                  ) -> escnn.group.Representation:
        r"""
        Band-Limited induced representation from the input representation ``repr`` up to frequency ``L`` (included).

        Args:
            L(int): max frequency
            repr (IrreducibleRepresentation): the representation of the subgroup
            subgroup_id: id identifying the subgroup H.

        """

        name = f"induced[{subgroup_id}][{repr.name}]_{L}"

        if name not in self.representations:
            subgroup, _, _ = self.subgroup(subgroup_id)

            homspace = self.homspace(subgroup_id)
            assert isinstance(repr, IrreducibleRepresentation)
            assert repr.group == subgroup

            irreps = []

            for l in range(L + 1):
                irr = self.irrep(l)
                multiplicity = homspace.dimension_basis(irr.id, repr.id)[1]
                irreps += [irr] * multiplicity

            assert len(irreps) > 0, f'Error! The induced representation from {repr.name} band-limited to {L} does not contain any irreps'

            self.representations[name] = directsum(irreps, name=name)

        return self.representations[name]

    def bl_irreps(self, L: int) -> List[Tuple]:
        r"""
        Returns a list containing the id of all irreps of frequency smaller or equal to ``L``.
        This method is useful to easily specify the irreps to be used to instantiate certain objects, e.g. the
        Fourier based non-linearity :class:`~escnn.nn.FourierPointwise`.
        """
        assert 0 <= L, L
        return [(l,) for l in range(L+1)]

    # @property
    def standard_representation(self) -> Representation:
        r"""
        Standard representation of :math:`\SO3` as 3x3 rotation matrices

        """
        name = f'standard'
    
        if name not in self._representations:
            
            change_of_basis = np.array([
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0]
            ])
            
            self._representations[name] = change_basis(
                self.irrep(1),
                change_of_basis=change_of_basis,
                name=name,
                supported_nonlinearities=self.irrep(1).supported_nonlinearities,
            )
    
        return self._representations[name]

    def irrep(self, l: int) -> IrreducibleRepresentation:
        r"""
        Build the irrep with rotational frequency :math:`l` of :math:`SO(3)`.
        Notice: the frequency has to be a non-negative integer.
        
        Args:
            l (int): the frequency of the irrep

        Returns:
            the corresponding irrep

        """
    
        assert l >= 0
        name = f"irrep_{l}"
        id = (l,)

        if id not in self._irreps:

            if l == 0:
                # Trivial representation
                irrep = build_trivial_irrep()
                character = build_trivial_character()
                supported_nonlinearities = ['pointwise', 'norm', 'gated', 'gate']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=0
                                                              )
            else:
        
                # other Irreducible Representations
                irrep = _build_irrep(l)
                character = _build_character(l)
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 2*l+1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=l)

        return self._irreps[id]

    def _clebsh_gordan_coeff(self, m, n, j) -> np.ndarray:
        group_keys = self._keys
        m = self.get_irrep_id(m)
        n = self.get_irrep_id(n)
        j = self.get_irrep_id(j)

        return _clebsh_gordan_tensor_so3(m[0], n[0], j[0])

    def _tensor_product_irreps(self, J: int, l: int) -> List[Tuple[Tuple, int]]:
        J, = self.get_irrep_id(J)
        l, = self.get_irrep_id(l)
        return [
            ((j,), 1)
            for j in range(np.abs(J - l), J + l + 1)
        ]

    _cached_group_instance = None

    @classmethod
    def _generator(cls, maximum_frequency: int = 3) -> 'SO3':
        if cls._cached_group_instance is None:
            cls._cached_group_instance = SO3(maximum_frequency)
        elif cls._cached_group_instance._maximum_frequency < maximum_frequency:
            cls._cached_group_instance._maximum_frequency = maximum_frequency
            cls._cached_group_instance._build_representations()
    
        return cls._cached_group_instance


def _clebsh_gordan_tensor_so3(m: int, n: int, j: int):
    if py3nj is None:
        return escnn.group._clebsh_gordan._clebsh_gordan_tensor((m,), (n,), (j,), SO3.__name__)
    else:
        m1 = np.arange(-m, m + 1).reshape(-1, 1, 1)
        m2 = np.arange(-n, n + 1).reshape(1, -1, 1)
        M = np.arange(-j, j + 1).reshape(1, 1, -1)
        _m = np.array([m]).reshape(1, 1, 1)
        _n = np.array([n]).reshape(1, 1, 1)
        _j = np.array([j]).reshape(1, 1, 1)
        cg = py3nj.clebsch_gordan(2 * _m, 2 * _n, 2 * _j, 2 * m1, 2 * m2, 2 * M)

        cob_m = _change_of_basis_real2complex(m)
        cob_n = _change_of_basis_real2complex(n)
        cob_j = _change_of_basis_real2complex(j).T.conj()

        cg = np.einsum('jlJ,KJ->jlK', cg, cob_j)
        cg = np.einsum('jlJ,jk->klJ', cg, cob_m)
        cg = np.einsum('jlJ,lk->jkJ', cg, cob_n)
        if (n + m + j) % 2 == 0:
            return cg.real.reshape(2 * m + 1, 2 * n + 1, 1, 2 * j + 1)
        else:
            return cg.imag.reshape(2 * m + 1, 2 * n + 1, 1, 2 * j + 1)


def _build_irrep(l: int):
    def irrep(e: GroupElement, l: int = l):
        return _wigner_d_matrix(e.value, l, param=e.param)
    
    return irrep


def _build_character(l: int):
    
    def character(e: GroupElement, l: int = l):
        return _character(e.value, l, param=e.param)
    
    return character

#############################################
# SUBGROUPS MAPS
#############################################


# SO(3) and Polyhedrons' Symmetries ###############################

def so3_to_rotgroup(adj: GroupElement, rotgroup: escnn.group.Group):
    assert isinstance(adj.group, SO3)
    assert 'Q' in rotgroup.PARAMETRIZATIONS
    
    def _map(e: GroupElement, rotgroup=rotgroup, adj=adj):
        so3 = adj.group
        assert e.group == so3
        try:
            return rotgroup.element(
                (adj @ e @ (~adj)).to('Q'),
                'Q'
            )
        except ValueError:
            return None

    return _map


def rotgroup_to_so3(adj: GroupElement, rotgroup: escnn.group.Group):
    assert isinstance(adj.group, SO3)
    assert 'Q' in rotgroup.PARAMETRIZATIONS

    def _map(e: GroupElement, rotgroup=rotgroup, adj=adj):
        assert e.group == rotgroup
        so3 = adj.group
        return (~adj) @ so3.element(e.to('Q'), 'Q') @ adj
    
    return _map


# SO(2) (and C_N) #####################################

def so3_to_so2(adj: GroupElement, so2: Union[escnn.group.SO2, escnn.group.CyclicGroup]):
    assert isinstance(adj.group, SO3)
    
    def _map(e: GroupElement, so2=so2, adj=adj):
        so3 = adj.group
        assert e.group == so3
        
        e = adj @ e @ (~adj)
        
        e = e.to('Q')
        
        if np.allclose(e[:2], 0.):
            # if the rotation is along the Z axis, i.e. on the XY plane
            s, c = e[2:]
            theta = 2 * np.arctan2(s, c)
            try:
                return so2.element(theta, 'radians')
            except ValueError:
                return None
        else:
            return None
        
    return _map


def so2_to_so3(adj: GroupElement, so2: Union[escnn.group.SO2, escnn.group.CyclicGroup]):
    assert isinstance(adj.group, SO3)
    
    def _map(e: GroupElement, so2=so2, adj=adj):
        assert e.group == so2
        so3 = adj.group
        
        theta_2 = e.to('radians') / 2.
        
        q = np.asarray([0., 0., np.sin(theta_2), np.cos(theta_2)])
        
        return (~adj) @ so3.element(q, 'Q') @ adj
    
    return _map


# O(2) (and D_N) ######################################

def so3_to_o2(adj: GroupElement, o2: Union[escnn.group.O2, escnn.group.DihedralGroup]):
    assert isinstance(adj.group, SO3)
    
    def _map(e: GroupElement, o2=o2, adj=adj):
        so3 = adj.group
        assert e.group == so3
        
        e = adj @ e @ (~adj)
        
        e = e.to('Q')
        
        if np.allclose(e[:2], 0.):
            # if it is a rotation along the Z axis, i.e. on the XY plane
            s, c = e[2:]
            theta = 2 * np.arctan2(s, c)
            flip = 0
            try:
                return o2.element((flip, theta), 'radians')
            except ValueError:
                return None
        elif np.allclose(e[2:], 0.):
            # if it is a rotation by 180 degrees around an axis perpendicular to Z, i.e. an axis in the XY plane
            c, s = e[:2]
            theta = 2 * np.arctan2(s, c)
            flip = 1
            try:
                return o2.element((flip, theta), 'radians')
            except ValueError:
                return None
        else:
            return None
    
    return _map


def o2_to_so3(adj: GroupElement, o2: Union[escnn.group.O2, escnn.group.DihedralGroup]):
    assert isinstance(adj.group, SO3)
    
    def _map(e: GroupElement, o2=o2, adj=adj):
        assert e.group == o2
        so3 = adj.group
        
        f, theta = e.to('radians')
        
        theta_2 = theta / 2.
        s, c = np.sin(theta_2), np.cos(theta_2)

        if f == 0:
            q = np.asarray([0., 0., s, c])
        elif f == 1:
            q = np.asarray([c, s, 0., 0.])
        else:
            raise ValueError()

        return (~adj) @ so3.element(q, 'Q') @ adj
    
    return _map


############################################


def check_parametrization():
    # check change of parametrizations
    
    parameterizations = (
        'Q',
        'MAT',
        'EV',
        'xyz', 'xzy', 'yxz', 'yzx', 'zxy', 'zyx', # Type-I Euler angles AKA Tait-Bryan angles
        'xyx', 'xzx', 'yxy', 'yzy', 'zxz', 'zyz',  # Type-II Euler angles
        
        'XYZ', 'XZY', 'YXZ', 'YZX', 'ZXY', 'ZYX',  # Type-I Euler angles AKA Tait-Bryan angles
        'XYX', 'XZX', 'YXY', 'YZY', 'ZXZ', 'ZYZ',  # Type-II Euler angles
    )
    
    G = SO3(2)
    
    er = set()
    
    for p2 in ['Q']:
        for p1 in parameterizations:
            for g in G.testing_elements(12):
                
                g_ = g._element
                g0 = _change_param(g_, p_from=PARAMETRIZATION, p_to=p1)
                g2 = _change_param(g0, p_from=p1, p_to=p2)
                g1 = _change_param(g2, p_from=p2, p_to=p1)
                
                if not G._equal(g0, g1, param1=p1, param2=p1):
                    print(g_)
                    print(g0)
                    print(g2)
                    print(g1)
                    print('-----------------------')
                    er.add((p1, p2))
                
                # assert G._equal(g0, g1, param1=p1, param2=p1, param=PARAMETRIZATION)#, (g, p1, p2)
    
    print(sorted(list(er)))
    
    for p1 in parameterizations:
        for g in G.testing_elements(5):
            g = g._element
            g0 = _change_param(g, p_from=PARAMETRIZATION, p_to=p1)
            g1 = _change_param(g0, p_from=p1, p_to='Q')
            
            assert np.allclose(np.linalg.norm(g1), 1.), (p1, g)
        

if __name__ == '__main__':
    check_parametrization()
    
    print(_change_param(np.array([0., 0., 0.]), 'zyz', 'Q'))
    
    for i in range(100):
        q = np.random.randn(4)
        q[np.random.randint(4)] = 0.
        q /= np.linalg.norm(q)

        m = _change_param(q, 'Q', 'MAT')
        q2 = _change_param(m, 'MAT', 'Q')
        np.set_printoptions(precision=3, suppress=True)
        if not (np.allclose(q, q2) or np.allclose(q, -q2)):
            print(q)
            print(q2)
            print('--------------')










from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
import torch.nn.functional as F

from typing import List, Tuple, Any

import numpy as np

__all__ = ["PointwiseNonLinearity"]


class PointwiseNonLinearity(EquivariantModule):
    
    def __init__(self, in_type: FieldType, function: str = 'p_relu'):
        r"""
        
        Pointwise non-linearities.
        The same scalar function is applied to every channel independently.
        The input representation is preserved by this operation and, therefore, it equals the output representation.
        
        Only representations supporting pointwise non-linearities are accepted as input field type.
        
        Args:
            in_type (FieldType):  the input field type
            function (str): the identifier of the non-linearity.
                    It is used to specify which function to apply.
                    By default (``'p_relu'``), ReLU is used.
            
        """

        assert isinstance(in_type.gspace, GSpace)
        
        super(PointwiseNonLinearity, self).__init__()

        for r in in_type.representations:
            assert 'pointwise' in r.supported_nonlinearities, \
                'Error! Representation "{}" does not support "pointwise" non-linearity'.format(r.name)
        
        self.space = in_type.gspace
        self.in_type = in_type
        
        # the representation in input is preserved
        self.out_type = in_type
        
        # retrieve the activation function to apply
        if function == 'p_relu':
            self._function = F.relu
        elif function == 'p_elu':
            self._function = F.elu
        elif function == 'p_sigmoid':
            self._function = F.sigmoid
        elif function == 'p_tanh':
            self._function = F.tanh
        else:
            raise ValueError('Function "{}" not recognized!'.format(function))

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Applies the pointwise activation function on the input fields

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map after the non-linearities have been applied

        """

        assert input.type == self.in_type
        
        # TODO - remove the 'contiguous()' call as soon as PyTorch's error is fixed
        # return GeometricTensor(self._function(input.tensor.contiguous()), self.out_type)
        return GeometricTensor(self._function(input.tensor), self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size
    
        x = torch.randn(3, c, 10, 10)
    
        x = GeometricTensor(x, self.in_type)
    
        errors = []
    
        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el)
            out2 = self(x.transform_fibers(el))
        
            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
        
            assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors




from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch.nn.functional as F
import torch

from typing import List, Tuple, Any, Union

import math

__all__ = [
    "PointwiseMaxPool2D", "PointwiseMaxPool",
    "PointwiseMaxPool3D",
    "PointwiseMaxPoolAntialiased2D", "PointwiseMaxPoolAntialiased",
    "PointwiseMaxPoolAntialiased3D",
]


class _PointwiseMaxPoolND(EquivariantModule):

    def __init__(self,
                 in_type: FieldType,
                 d: int,
                 kernel_size: Union[int, Tuple],
                 stride: Union[int, Tuple] = None,
                 padding: Union[int, Tuple] = 0,
                 dilation: Union[int, Tuple] = 1,
                 ceil_mode: bool = False
                 ):
        r"""

        Channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.MaxPool2D` or :class:`torch.nn.MaxPool3D`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            d (int): dimensionality of the base space (2 for images, 3 for volumes)
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when True, will use ceil instead of floor to compute the output shape

        """

        assert d in [2, 3], f"Only dimensionality 2 or 3 are currently suported by 'd={d}' found"

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == d, (in_type.gspace.dimensionality, d)

        for r in in_type.representations:
            assert 'pointwise' in r.supported_nonlinearities, \
                f"""Error! Representation "{r.name}" does not support pointwise non-linearities
                so it is not possible to pool each channel independently"""

        super(_PointwiseMaxPoolND, self).__init__()

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        if isinstance(kernel_size, int):
            self.kernel_size = (kernel_size,) * self.d
        else:
            self.kernel_size = kernel_size

        if isinstance(stride, int):
            self.stride = (stride,) * self.d
        elif stride is None:
            self.stride = self.kernel_size
        else:
            self.stride = stride

        if isinstance(padding, int):
            self.padding = (padding,) * self.d
        else:
            self.padding = padding

        if isinstance(dilation, int):
            self.dilation = (dilation,) * self.d
        else:
            self.dilation = dilation

        self.ceil_mode = ceil_mode

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map

        """

        assert input.type == self.in_type

        # run the common max-pooling
        if self.d == 2:
            output = F.max_pool2d(input.tensor,
                                  self.kernel_size,
                                  self.stride,
                                  self.padding,
                                  self.dilation,
                                  self.ceil_mode)
        elif self.d == 3:
            output = F.max_pool3d(input.tensor,
                                  self.kernel_size,
                                  self.stride,
                                  self.padding,
                                  self.dilation,
                                  self.ceil_mode)
        else:
            raise NotImplementedError

        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        assert len(input_shape) == 2 + self.d, (input_shape, self.d)
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        shape_i = input_shape[2:]

        # compute the output shape (see 'torch.nn.MaxPool2D')
        shape_o = tuple(
            (shape_i[i] + 2 * self.padding[i] - self.dilation[i] * (self.kernel_size[i] - 1) - 1) / self.stride[i] + 1
            for i in range(self.d)
        )

        if self.ceil_mode:
            shape_o = tuple(
                math.ceil(shape_o[i])
                for i in range(self.d)
            )
        else:
            shape_o = tuple(
                math.floor(shape_o[i])
                for i in range(self.d)
            )

        return (b, self.out_type.size, *shape_o)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:

        # this kind of pooling is not really equivariant so we can not test equivariance
        pass

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.MaxPool2d` module and set to "eval" mode.

        """

        self.eval()

        if self.d == 2:
            return torch.nn.MaxPool2d(self.kernel_size, self.stride, self.padding, self.dilation).eval()
        elif self.d == 3:
            return torch.nn.MaxPool2d(self.kernel_size, self.stride, self.padding, self.dilation).eval()
        else:
            raise NotImplementedError


class _PointwiseMaxPoolAntialiasedND(_PointwiseMaxPoolND):

    def __init__(self,
                 in_type: FieldType,
                 d: int,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 dilation: Union[int, Tuple[int, int]] = 1,
                 ceil_mode: bool = False,
                 sigma: float = 0.6,
                 ):
        r"""

        Anti-aliased version of channel-wise max-pooling (each channel is treated independently).

        The max over a neighborhood is performed pointwise withot downsampling.
        Then, convolution with a gaussian blurring filter is performed before downsampling the feature map.

        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.


        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            d (int): dimensionality of the base space (2 for images, 3 for volumes)
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape
            sigma (float): standard deviation for the Gaussian blur filter

        """

        if dilation != 1:
            raise NotImplementedError("Dilation larger than 1 is not supported yet")

        super(_PointwiseMaxPoolAntialiasedND, self).__init__(in_type, d, kernel_size, stride, padding, dilation, ceil_mode)

        assert sigma > 0.

        filter_size = 2 * int(round(4 * sigma)) + 1

        # Build the Gaussian smoothing filter

        grid = torch.stack(torch.meshgrid(*[torch.arange(filter_size)] * self.d, indexing='ij'), dim=-1)

        mean = (filter_size - 1) / 2.
        variance = sigma ** 2.

        # setting the dtype is needed, otherwise it becomes an integer tensor
        r = -torch.sum((grid - mean) ** 2., dim=-1, dtype=torch.get_default_dtype())

        # Build the gaussian kernel
        _filter = torch.exp(r / (2 * variance))

        # Normalize
        _filter /= torch.sum(_filter)

        # The filter needs to be reshaped to be used in 2d depthwise convolution
        _filter = _filter.view(1, 1, *[filter_size]*self.d).repeat((in_type.size, 1, *[1]*self.d))

        self.register_buffer('filter', _filter)
        self._pad = tuple(p + int((filter_size - 1) // 2) for p in self.padding)

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map

        """

        assert input.type == self.in_type

        if self.d == 2:
            # evaluate the max operation densely (stride = 1)
            output = F.max_pool2d(input.tensor,
                                  self.kernel_size,
                                  1,
                                  self.padding,
                                  self.dilation,
                                  self.ceil_mode)
            output = F.conv2d(output, self.filter, stride=self.stride, padding=self._pad, groups=output.shape[1])
        elif self.d == 3:
            # evaluate the max operation densely (stride = 1)
            output = F.max_pool3d(input.tensor,
                                  self.kernel_size,
                                  1,
                                  self.padding,
                                  self.dilation,
                                  self.ceil_mode)
            output = F.conv3d(output, self.filter, stride=self.stride, padding=self._pad, groups=output.shape[1])
        else:
            raise NotImplementedError

        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords=None)

    def export(self):
        raise NotImplementedError


class PointwiseMaxPool2D(_PointwiseMaxPoolND):

    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 dilation: Union[int, Tuple[int, int]] = 1,
                 ceil_mode: bool = False
                 ):
        r"""

        Channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.MaxPool2D`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.
        
        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when True, will use ceil instead of floor to compute the output shape

        """

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == 2

        super(PointwiseMaxPool2D, self).__init__(
            in_type, 2, kernel_size, stride, padding, dilation, ceil_mode
        )


class PointwiseMaxPool3D(_PointwiseMaxPoolND):

    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 dilation: Union[int, Tuple[int, int]] = 1,
                 ceil_mode: bool = False
                 ):
        r"""

        Channel-wise max-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.MaxPool3D`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.

        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when True, will use ceil instead of floor to compute the output shape

        """

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == 3

        super(PointwiseMaxPool3D, self).__init__(
            in_type, 3, kernel_size, stride, padding, dilation, ceil_mode
        )


class PointwiseMaxPoolAntialiased2D(_PointwiseMaxPoolAntialiasedND):

    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 dilation: Union[int, Tuple[int, int]] = 1,
                 ceil_mode: bool = False,
                 sigma: float = 0.6,
                 ):
        r"""

        Anti-aliased version of channel-wise max-pooling (each channel is treated independently).

        The max over a neighborhood is performed pointwise withot downsampling.
        Then, convolution with a gaussian blurring filter is performed before downsampling the feature map.

        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.


        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape
            sigma (float): standard deviation for the Gaussian blur filter

        """

        super(PointwiseMaxPoolAntialiased2D, self).__init__(
            in_type, 2, kernel_size, stride, padding, dilation, ceil_mode, sigma
        )


class PointwiseMaxPoolAntialiased3D(_PointwiseMaxPoolAntialiasedND):

    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 dilation: Union[int, Tuple[int, int]] = 1,
                 ceil_mode: bool = False,
                 sigma: float = 0.6,
                 ):
        r"""

        Anti-aliased version of channel-wise max-pooling (each channel is treated independently).

        The max over a neighborhood is performed pointwise withot downsampling.
        Then, convolution with a gaussian blurring filter is performed before downsampling the feature map.

        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.


        Notice that not all representations support this kind of pooling. In general, only representations which support
        pointwise non-linearities do.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape
            sigma (float): standard deviation for the Gaussian blur filter

        """

        super(PointwiseMaxPoolAntialiased3D, self).__init__(
            in_type, 3, kernel_size, stride, padding, dilation, ceil_mode, sigma
        )


# for backward compatibility
PointwiseMaxPool = PointwiseMaxPool2D
PointwiseMaxPoolAntialiased = PointwiseMaxPoolAntialiased2D







from typing import List, Tuple, Any

import numpy as np

from collections import defaultdict

from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch

from torch.nn import Parameter

from .gated1 import GATED_ID, GATES_ID

__all__ = ["GatedNonLinearity2"]


class GatedNonLinearity2(EquivariantModule):
    
    def __init__(self,
                 in_type: Tuple[FieldType, FieldType],
                 **kwargs
                 ):
        r"""
        
        Gated non-linearities.
        This module applies a bias and a sigmoid function of the gates fields and, then, multiplies each gated
        field by one of the gates.
        
        The input representation of the gated fields is preserved by this operation while the gate fields are
        discarded.
        
        The gates and the gated fields are provided in two different tensors: :attr:`in_repr` is a tuple containing two
        representations: the representation of the tensor containing only the gates (which have to be trivial fields)
        and the representation of the tensor containing only the gated fields. Therefore, two tensors need to be
        provided as input to the ``forward`` method: the first contains the gates and the second the gated fields.
        Finally, notice that the number of gates and the number of gated fields have to match, i.e. the two
        representations need to have the same number of fields.
        
        .. todo ::
            This module has 2 input tensors and, so, two input field types.
            EquivariantModule only supports one input though.
            Fix this.
        
        Args:
            in_type (Tuple): a pair containing, in order, the field type of the gates and the field type
                             of the gated fields
            
        """

        assert len(in_type) == 2
        assert isinstance(in_type[0], FieldType)
        assert isinstance(in_type[1], FieldType)
        
        assert isinstance(in_type[0].gspace, GSpace)
        assert in_type[0].gspace == in_type[1].gspace

        # the first is the representation of the gates
        for r in in_type[0].representations:
            # assert GATES_ID in r.supported_nonlinearities, \
            assert r.is_trivial(), \
                "Error! Representation \"{}\" can't be a \"gate\"".format(r.name)

        # the second is the representation of the gated fields
        for r in in_type[1].representations:
            assert GATED_ID in r.supported_nonlinearities, \
                'Error! Representation "{}" does not support "gated" non-linearity'.format(r.name)

        assert len(in_type[0]) == len(in_type[1]), \
            'Error! Number of gates ({}) does not match the number of gated non-linearities required ({})' \
                .format(len(in_type[0]), len(in_type[1]))
        
        super(GatedNonLinearity2, self).__init__()

        self.space = in_type[0].gspace
        
        self._contiguous = {}
        _input_indices = defaultdict(lambda: [])
        
        self._nfields = defaultdict(int)
        
        self.branching = None

        self.in_type = in_type
        self.out_type = in_type[1]
        
        ngates = len(in_type[0])
        
        # group fields by their size, check if fields with the same size are consecutive
        # and retrieve the indices of the fields
        last_position = 0
        last_size = None
        for r in in_type[1].representations:

            if r.size != last_size:
                if not r.size in self._contiguous:
                    self._contiguous[r.size] = True
                else:
                    self._contiguous[r.size] = False
            last_size = r.size
            
            self._nfields[r.size] += 1
            _input_indices[r.size] += list(range(last_position, last_position + r.size))
            last_position += r.size

        _input_indices = dict(_input_indices)
        
        for s, contiguous in self._contiguous.items():
            if contiguous:
                # if the fields are contiguous, only the first and last indices are kept
                _input_indices[s] = torch.LongTensor([min(_input_indices[s]), max(_input_indices[s])+1])
            else:
                # otherwise, transform the list of indices into a tensor
                _input_indices[s] = torch.LongTensor(_input_indices[s])
                
            # register the indices tensors as parameters of this module
            self.register_buffer('indices_{}'.format(s), _input_indices[s])
        
        # the gated fields are preserved and, therefore, the output representation is the same
        # as the gated fields input representation
        
        # build a sorted list of the fields groups, such that every time they are iterated through in the same order
        self._order = sorted(_input_indices.keys())
        
        # the bias for the gates
        self.bias = Parameter(torch.randn(1, ngates, 1, dtype=torch.float), requires_grad=True)

    def forward(self, gates: GeometricTensor, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Apply the gated non-linearity to the input feature map.
        
        Args:
            gates (GeometricTensor): feature map corresponding to the gates
            input (GeometricTensor): input feature map corresponding to the gated fields

        Returns:
            the resulting feature map
            
        """
        
        assert isinstance(input, GeometricTensor)
        assert isinstance(gates, GeometricTensor)
        assert gates.type == self.in_type[0]
        assert input.type == self.in_type[1]
        
        assert gates.has_same_coords(input)

        coords = input.coords
        
        gates = gates.tensor
        input = input.tensor
        
        assert gates.shape[0] == input.shape[0]
        assert gates.shape[2:] == input.shape[2:]

        b, c = input.shape[:2]
        spatial_dims = input.shape[2:]

        # transform the gates
        gates = torch.sigmoid(
            gates -
            self.bias.view(1, -1, *[1]*len(spatial_dims))
        )

        # build the output tensor
        output = torch.empty(b, self.out_type.size, *spatial_dims, dtype=torch.float, device=self.bias.device)

        next_gate = 0
        
        # for each field size
        for size in self._order:
            
            indices = getattr(self, f"indices_{size}")
            
            # retrieve the needed gates
            g = gates[:, next_gate:next_gate + self._nfields[size], ...].view(b, -1, 1, *spatial_dims)
            
            if self._contiguous[size]:
                # if the fields were contiguous, we can use slicing
                output[:, indices[0]:indices[1], ...] =\
                    (
                            input[:, indices[0]:indices[1], ...]
                            .view(b, -1, size, *spatial_dims)
                            * g
                    ).view(b, -1, *spatial_dims)
            
            else:
                # otherwise we have to use indexing
                output[:, indices, ...] = \
                    (
                            input[:, indices, ...]
                            .view(b, -1, size, *spatial_dims)
                            * g
                    ).view(b, -1, *spatial_dims)
            
            # shift the position on the gates fiber
            next_gate += self._nfields[size]
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, *input_shapes: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        assert len(input_shapes) == 2
        gates_shape, input_shape = input_shapes
        
        assert len(gates_shape) == 4
        assert gates_shape[1] == self.in_type[0].size

        assert len(input_shape) == 4
        assert input_shape[1] == self.in_type[1].size
        
        assert gates_shape[0] == input_shape[0]
        assert gates_shape[2:] == input_shape[2:]

        b, c, hi, wi = input_shape

        return b, self.out_type.size, hi, wi
        
    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = [self.in_type[i].size for i in range(2)]

        x = [torch.randn(3, c[i], 10, 10) for i in range(2)]

        x = [GeometricTensor(x[i], self.in_type[i]) for i in range(2)]

        errors = []

        for el in self.space.testing_elements:
            
            out1 = self(*x).transform_fibers(el)
            out2 = self(*[x[i].transform_fibers(el) for i in range(2)])
    
            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
    
            assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
    
            errors.append((el, errs.mean()))

        return errors





from __future__ import annotations

import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import connected_components

from typing import Tuple, List


import os
from joblib import Memory

import escnn
from escnn.group._numerical import find_intertwiner_basis_sylvester, build_sylvester_constraint
from escnn.group._numerical import InsufficientIrrepsException

from escnn.group import __cache_path__
cache = Memory(__cache_path__, verbose=0)


########################################################################################################################
# Numeric solutions for CG coeffs
########################################################################################################################

class UnderconstrainedCGSystem(Exception):
    
    def __init__(
            self, 
            G: escnn.group.Group,
            J: Tuple,
            l: Tuple,
            j: Tuple,
            S: int,
            message: str = 'The algorithm to compute the CG coefficients failed due to an unsufficient number of samples to constraint the problem',
    ):
        self.G = G 
        self.J = J 
        self.l = l 
        self.j = j 
        self.S = S 
        super(UnderconstrainedCGSystem, self).__init__(message)


MAX_SAMPLES = 20


# cache_cg = Memory(os.path.join(os.path.dirname(__file__), '_jl_clebshgordan'), verbose=2)


# @cache_cg.cache
@cache.cache
def _clebsh_gordan_tensor(J: Tuple, l: Tuple, j: Tuple, group_class: str, **group_keys) -> np.ndarray:
    
    G = escnn.group.groups_dict[group_class]._generator(**group_keys)
    
    psi_J = G.irrep(*J)
    psi_l = G.irrep(*l)
    psi_j = G.irrep(*j)
    
    D = psi_J.size * psi_l.size * psi_j.size
    
    def build_matrices(samples):
        D_Jl = []
        D_j = []
        for g in samples:
            D_J_g = psi_J(g)
            D_l_g = psi_l(g)
            D_j_g = psi_j(g)
        
            D_Jl_g = np.kron(D_J_g, D_l_g)
            
            D_j.append(D_j_g)
            D_Jl.append(D_Jl_g)
        return D_Jl, D_j
    
    try:
        generators = G.generators
        S = len(generators)
    except ValueError:
        generators = []
        # number of samples to use to approximate the solutions
        # usually 3 are sufficient
        S = 3

    _S = S
    
    while True:
        # sometimes it might not converge, so we need to try a few times
        attepts = 5
        while True:
            try:
                samples = generators + [G.sample() for _ in range(S - len(generators))]
                if len(samples) == 0:
                    basis = np.eye(D)
                else:
                    D_Jl, D_j = build_matrices(samples)
                    basis = find_intertwiner_basis_sylvester(D_Jl, D_j)
                
            except np.linalg.LinAlgError:
                if attepts > 0:
                    attepts -= 1
                    continue
                else:
                    raise
            else:
                break
                
        # check that the solutions found are also in the kernel of the constraint matrix built with other random samples
        D_Jl, D_j = build_matrices(generators + [G.sample() for _ in range(20)])
        tensor = build_sylvester_constraint(D_Jl, D_j).todense().reshape(-1, D)
        
        if np.allclose(tensor @ basis, 0.):
            break
        elif S < MAX_SAMPLES:
            # if this not the case, try again using more samples to build the constraint matrix
            S += 1
        else:
            raise UnderconstrainedCGSystem(G, psi_J.id, psi_l.id, psi_j.id, S)
    
    if S > _S:
        print(G.name, psi_J.id, psi_l.id, psi_j.id, S)

    # the dimensionality of this basis corresponds to the multiplicity of `j` in the tensor-product `J x l`
    s = basis.shape[1]
    assert s % psi_j.sum_of_squares_constituents == 0

    jJl = s // psi_j.sum_of_squares_constituents

    CG = basis.reshape((psi_j.size, psi_J.size, psi_l.size, s)).transpose(1, 2, 3, 0)
    # CG indexed as [J, l, s, j]

    if s == 0:
        return CG

    norm = np.sqrt((CG**2).mean(2, keepdims=True).sum(1, keepdims=True).sum(0, keepdims=True))
    CG /= norm
    
    ortho = np.einsum(
        'Jlsj,Jlti,kji->stk',
        CG, CG, psi_j.endomorphism_basis()
    )
    
    ortho = (ortho**2).sum(2) > 1e-9
    assert ortho.astype(np.uint).sum() == s * psi_j.sum_of_squares_constituents, (ortho, s, jJl, psi_j.sum_of_squares_constituents)

    n, dependencies = connected_components(csgraph=csr_matrix(ortho), directed=False, return_labels=True)
    assert n * psi_j.sum_of_squares_constituents == s, (ortho, n, s, psi_j.sum_of_squares_constituents)

    mask = np.zeros((ortho.shape[0]), dtype=bool)
    for i in range(n):
        columns = np.nonzero(dependencies == i)[0]
        assert len(columns) == psi_j.sum_of_squares_constituents
        selected_column = columns[0]
        mask[selected_column] = 1

    assert mask.sum() == n

    CG = CG[..., mask, :]

    assert CG.shape[-2] == jJl

    B = CG.reshape(-1, psi_j.size * jJl)
    assert np.allclose(B.T@B, np.eye(psi_j.size * jJl))
    
    return CG


# cache_dec = Memory(os.path.join(os.path.dirname(__file__), '_jl_tensor_decomposition'), verbose=2)

# @cache_dec.cache
@cache.cache
def _find_tensor_decomposition(J: Tuple, l: Tuple, group_class: str, **group_keys) -> List[Tuple[Tuple, int]]:
    G = escnn.group.groups_dict[group_class]._generator(**group_keys)
    
    psi_J = G.irrep(*J)
    psi_l = G.irrep(*l)
    
    irreps = []
    
    size = 0
    for psi_j in G.irreps():
        CG = G._clebsh_gordan_coeff(psi_J, psi_l, psi_j)
        
        S = CG.shape[-2]
        
        if S > 0:
            irreps.append((psi_j.id, S))
        
        size += psi_j.size * S

    # check that size == psi_J.size * psi_l.size
    
    if size < psi_J.size * psi_l.size:
        from textwrap import dedent
        message = dedent(f"""
            Error! Did not find sufficient irreps to complete the decomposition of the tensor product of '{psi_J.name}' and '{psi_l.name}'.
            It is likely this happened because not sufficiently many irreps in '{G}' have been instantiated.
            Try instantiating more irreps and then repeat this call.
            The sum of the sizes of the irreps found is {size}, but the representation has size {psi_J.size * psi_l.size}.
        """)
        raise InsufficientIrrepsException(G, message)

    assert size <= psi_J.size * psi_l.size, f"""
        Error! Found too many irreps in the the decomposition of the tensor product of '{psi_J.name}' and '{psi_l.name}'.
        This should never happen!
    """

    return irreps


########################################################################################################################
# UNIT TESTS
########################################################################################################################


def check_spherical_harmonics():
    # check compatibility with SH

    from escnn.group import SO3
    from escnn.group.groups.so3group import spherical_harmonic
    
    G = SO3(2)
    
    def cart_to_sp(point):
        x, y, z = point.reshape(3)
        
        theta = np.arccos(np.clip(z, -1., 1.))
        phi = np.arctan2(y, x)
        return np.asarray([theta, phi])
    
    def sp_to_cart(angles):
        x = np.sin(angles[0]) * np.cos(angles[1])
        y = np.sin(angles[0]) * np.sin(angles[1])
        z = np.cos(angles[0])
        return np.asarray([x, y, z]).reshape(-1, 1)
    
    g = np.array([np.pi / 2, 0., 0.])
    print(G.element(g, 'zyz').to('MAT'))
    
    for l in range(9):
        
        print(l)
        n = 3
        for a in range(n):
            phi = 2 * np.pi * a / n
            for b in range(n):
                # theta = np.pi * (2 * b + 1) / (2*n)
                # theta = np.pi * (b + 1) / (n+2)
                # sample around b = 0  and b = np.pi to check singularities
                theta = np.pi * b / n
                
                alpha = np.asarray([theta, phi])
                x = sp_to_cart(alpha)
                
                Yx = spherical_harmonic(l, alpha.reshape(2, 1))
                
                for g in G.testing_elements():
                    # gx = G.irrep(1)(g) @ x
                    gx = G.standard_representation()(g) @ x
                    # gx = G.change_param(g, PARAMETRIZATION, 'MAT') @ x
                    galpha = cart_to_sp(gx)
                    
                    Ygx = spherical_harmonic(l, galpha.reshape(2, 1))
                    
                    gYx = G.irrep(l)(g) @ Yx
                    # print("###################################")
                    # print(g, gx.flatten(), galpha)
                    # print(Yx)
                    # print(Ygx)
                    # print(G.irrep(l)(g) @ Yx)
                    
                    assert np.allclose(gYx, Ygx, atol=1e-6, rtol=1e-6), g


def check_spherical_harmonics2():
    # check compatibility of CG coeffs with SH
    
    from escnn.group import SO3
    from escnn.group.groups.so3group import spherical_harmonic
    
    G = SO3(2)

    def cart_to_sp(point):
        x, y, z = point.reshape(3)
    
        theta = np.arccos(np.clip(z, -1., 1.))
        phi = np.arctan2(y, x)
        return np.asarray([theta, phi])

    def sp_to_cart(angles):
        x = np.sin(angles[0]) * np.cos(angles[1])
        y = np.sin(angles[0]) * np.sin(angles[1])
        z = np.cos(angles[0])
        return np.asarray([x, y, z]).reshape(-1, 1)
    
    for J in range(6):
        for l in range(6):
            for j in range(np.abs(J-l), J+l):
                
                CG = G._clebsh_gordan_coeff(l, J, j).reshape(2*l+1, 2*J+1, 2*j+1)
                
                print(l, J, j)
                
                n = 3
                for a in range(n):
                    phi = 2 * np.pi * a / n
                    for b in range(n):
                        theta = np.pi * b / n
                        
                        alpha = np.asarray([theta, phi])
                        x = sp_to_cart(alpha)
                        
                        Yx = spherical_harmonic(j, alpha.reshape(2, 1))
                        Yx = np.einsum('lJj,jp->lJp', CG, Yx)
                        
                        for g in G.testing_elements():
                            gx = G.standard_representation()(g) @ x
                            galpha = cart_to_sp(gx)
                            
                            Ygx = spherical_harmonic(j, galpha.reshape(2, 1))
                            Ygx = np.einsum('lJj,jp->lJp', CG, Ygx)

                            gYx = np.einsum('Ll,ljp,jJ->LJp', G.irrep(l)(g), Yx, G.irrep(J)(g).T)
                            assert np.allclose(gYx, Ygx, atol=1e-6, rtol=1e-6), g


def test_kron():
    
    for _ in range(5):
        d1 = 5
        d2 = 3
        A = np.random.randn(d1, d1)
        B = np.random.randn(d2, d2)
        
        x = np.random.randn(d1, d2)
        
        y1 = A @ x @ B
        
        y2 = np.kron(A, B.T) @ x.reshape(-1, 1)
        y2 = y2.reshape(d1, d2)
        
        assert np.allclose(y1, y2)


if __name__ == '__main__':
    # check_spherical_harmonics()
    check_spherical_harmonics2()
    # test_kron()





from escnn.group import *
from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from escnn.nn import init

from ..equivariant_module import EquivariantModule

import torch
from escnn.nn.modules.basismanager import BasisManager
from escnn.nn.modules.basismanager import BlocksBasisExpansion

from torch.nn import Parameter
from torch.nn.functional import linear

from typing import List, Tuple, Any

import numpy as np

__all__ = ['TensorProductModule']


class TensorProductModule(EquivariantModule):

    def __init__(self, in_type: FieldType, out_type: FieldType, initialize: bool = True):
        r"""

        Applies a (learnable) quadratic non-linearity to the input features.

        The module requires its input and output types to be *uniform*, i.e. contain multiple copies of the same
        representation; see also :meth:`~escnn.nn.FieldType.uniform`.
        Moreover, the input and output field types must have the same number of fields, i.e.
        ```len(in_type) == len(out_type)```.

        The module computes the tensor product of each field with itself to generate an intermediate feature map.
        Note that this feature map will have size ```len(in_type) * in_type.representations[0].size**2```.
        To prevent the exponential growth of the model's width at each layer, the module includes also a learnable
        linear projection of each ```in_type.representations[0].size**2```-dimensional output field to a corresponding
        ```out_type.representations[0].size``` output field.
        Note that this layer applies an independent linear projection to each field individually but does not mix them.

        ..warning ::
            A model employing only this kind of non-linearities will effectively be a polynomial function.
            Moreover, the degree of the polynomial grows exponentially with the depth of the network.
            This may result in some instabilities during training.

        Args:
            in_type (FieldType): the type of the input field, specifying its transformation law
            out_type (FieldType): the type of the output field, specifying its transformation law
            initialize (bool, optional): initialize the weights of the model (warning: can be slow). Default: ``True``

        Attributes:
            ~.weights (torch.Tensor): the learnable parameters which are used to expand the projection matrix
            ~.matrix (torch.Tensor): the matrix obtained by expanding the parameters in :attr:`~escnn.nn.TensorProductModule.weights`

        """
        super(TensorProductModule, self).__init__()

        self.in_type = in_type
        self.out_type = out_type

        assert self.in_type.gspace == self.out_type.gspace

        self.gspace: GSpace = self.in_type.gspace

        assert in_type.uniform
        assert out_type.uniform
        assert len(in_type) == len(out_type), (len(in_type), len(out_type))

        # number of independent fields
        self._C = len(in_type)

        in_repr = in_type.representations[0]
        hid_repr = in_repr.tensor(in_repr)

        self.hidden_type = self.gspace.type(*[hid_repr]*self._C)

        # BlocksBasisExpansion: submodule which takes care of building the linear projection matrix
        self._basisexpansion = BlocksBasisExpansion(
            # we use only the first representation in input since we want a "grouped" linear layer,not a dense one
            [hid_repr], out_type.representations,
            self.gspace.build_fiber_intertwiner_basis,
            np.zeros((1, 1)),
            recompute=False
        )

        if self.basisexpansion.dimension() == 0:
            raise ValueError('''
                The basis for the steerable matrix is empty!
            ''')

        self.weights = Parameter(torch.zeros(self.basisexpansion.dimension()), requires_grad=True)
        self.weights.data[:] = torch.randn_like(self.weights)
        if initialize:
            # by default, the weights are initialized with a generalized form of He's weight initialization
            init.generalized_he_init(self.weights.data, self.basisexpansion)

        matrix_size = (out_type.size, hid_repr.size)
        self.register_buffer("matrix", torch.zeros(*matrix_size))

    @classmethod
    def construct(cls, gspace: GSpace, channels: int, in_repr: Representation, out_repr: Representation, initialize: bool = True) -> 'TensorProductModule':
        f'''
            Constructor method which provides an alternative way to instantiate a TensorProductModule.
            This method automatically builds the ```in_type``` (respectively, ```out_type```) as ```channels``` copies 
            of ```in_repr``` (```out_repr```).
        '''
        assert in_repr.group == gspace.fibergroup
        assert out_repr.group == gspace.fibergroup
        in_type = gspace.type(*[in_repr]*channels)
        out_type = gspace.type(*[out_repr]*channels)
        return TensorProductModule(in_type, out_type, initialize=initialize)

    def forward(self, input: GeometricTensor):

        assert input.type == self.in_type

        coords = input.coords
        spatial_shape = input.shape[2:]
        input = input.tensor.view(input.shape[0], self._C, self.in_type.representations[0].size,  *spatial_shape)

        # compute tensor product
        tensor_features = torch.einsum('bci...,bco...->bcio...', input, input)
        tensor_features = tensor_features.view(input.shape[0], self._C, self.hidden_type.representations[0].size, *spatial_shape)

        # perform projection
        if not self.training:
            _matrix = self.matrix
        else:
            # retrieve the matrix and the bias
            _matrix = self.expand_parameters()

        _matrix = _matrix.view(self._C, self.out_type.representations[0].size, self.hidden_type.representations[0].size)
        output = torch.einsum('coi,bci...->bco...', _matrix, tensor_features)

        output = output.view(input.shape[0], self.out_type.size, *spatial_shape)

        return self.out_type(output, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        if self.in_type != self.out_type:
            return input_shape[:1] + (self.out_type.size, ) + input_shape[2:]
        else:
            return input_shape

    def export(self):
        raise NotImplementedError()

    @property
    def basisexpansion(self) -> BasisManager:
        r"""
        Submodule which takes care of building the matrix.

        It uses the learnt ``weights`` to expand a basis and returns a matrix in the usual form used by conventional
        convolutional modules.
        It uses the learned ``weights`` to expand the kernel in the G-steerable basis and returns it in the shape
        :math:`(c_\text{out}, c_\text{in}, s^d)`, where :math:`s` is the ``kernel_size`` and :math:`d` is the
        dimensionality of the base space.

        """
        return self._basisexpansion

    def expand_parameters(self) -> torch.Tensor:
        r"""

        Expand the matrix in terms of the :attr:`escnn.nn.TensorProductModule.weights`.

        Returns:
            the expanded projection matrix

        """
        _matrix = self.basisexpansion(self.weights)
        return _matrix.view(self.out_type.size, self.hidden_type.representations[0].size)

    def train(self, mode=True):
        r"""

        If ``mode=True``, the method sets the module in training mode and discards the
        :attr:`~escnn.nn.TensorProductModule.matrix` attribute.

        If ``mode=False``, it sets the module in evaluation mode. Moreover, the method builds the matrix
        using the current values of the trainable parameters and store them in :attr:`~escnn.nn.TensorProductModule.matrix`
        such that it is not recomputed at each forward pass.

        .. warning ::

            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of this class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.

        Args:
            mode (bool, optional): whether to set training mode (``True``) or evaluation mode (``False``).
                                   Default: ``True``.

        """

        if not mode:
            _matrix = self.expand_parameters()
            self.register_buffer("matrix", _matrix)
        else:
            if hasattr(self, "matrix"):
                del self.matrix

        return super(TensorProductModule, self).train(mode)





from escnn.group import Representation
from escnn.kernels import KernelBasis, EmptyBasisException


from escnn.nn.modules.basismanager import retrieve_indices
from .basismanager import BasisManager

from escnn.nn.modules.basismanager.basissampler_singleblock import block_basissampler

from typing import Callable, Tuple, Dict, List, Iterable, Union
from collections import defaultdict

import torch
import numpy as np
import math


__all__ = ["BlocksBasisSampler"]


class BlocksBasisSampler(torch.nn.Module, BasisManager):

    def __init__(self,
                 in_reprs: List[Representation],
                 out_reprs: List[Representation],
                 basis_generator: Callable[[Representation, Representation], KernelBasis],
                 basis_filter: Callable[[dict], bool] = None,
                 recompute: bool = False,
                 ):
        r"""

        Module which performs the expansion of an analytical filter basis and samples it on arbitrary input points.

        Args:
            in_reprs (list): the input field type
            out_reprs (list): the output field type
            basis_generator (callable): method that generates the analytical filter basis
            basis_filter (callable, optional): filter for the basis elements. Should take a dictionary containing an
                                               element's attributes and return whether to keep it or not.
            recompute (bool, optional): whether to recompute new bases or reuse, if possible, already built tensors.

        """
    
        super(BlocksBasisSampler, self).__init__()

        self._in_reprs = in_reprs
        self._out_reprs = out_reprs
        self._input_size = sum(r.size for r in in_reprs)
        self._output_size = sum(r.size for r in out_reprs)

        self._in_sizes = {
            r.name: r.size for r in in_reprs
        }

        uniform_input = all(r == in_reprs[0] for r in in_reprs)
        uniform_output = all(r == out_reprs[0] for r in out_reprs)
        _uniform = uniform_input and uniform_output

        # we group the bases by their input and output representations
        _block_sampler_modules = {}

        # iterate through all different pairs of input/output representations
        # and, for each of them, build a basis
        for i_repr in set(in_reprs):
            for o_repr in set(out_reprs):
                reprs_names = (i_repr.name, o_repr.name)
                try:

                    basis = basis_generator(i_repr, o_repr)

                    # BasisSamplerSingleBlock: sampler for block with input i_repr and output o_repr
                    block_sampler = block_basissampler(basis, basis_filter=basis_filter, recompute=recompute)
                    _block_sampler_modules[reprs_names] = block_sampler

                    # register the block sampler as a submodule
                    self.add_module(f"block_sampler_{reprs_names}", block_sampler)

                except EmptyBasisException:
                    # print(f"Empty basis at {reprs_names}")
                    pass

        if len(_block_sampler_modules) == 0:
            print('WARNING! The basis for the block sampler of the filter is empty!')

        # the list of all pairs of input/output representations which don't have an empty basis
        self._representations_pairs = sorted(list(_block_sampler_modules.keys()))

        self._n_pairs = len(set(in_reprs)) * len(set(out_reprs))
        if _uniform:
            assert self._n_pairs <= 1
        self._uniform = _uniform and self._n_pairs == 1

        # retrieve for each representation in both input and output fields:
        # - the number of its occurrences,
        # - the indices where it occurs and
        # - whether its occurrences are contiguous or not
        self._in_count, _in_indices, _in_contiguous = retrieve_indices(in_reprs)
        self._out_count, _out_indices, _out_contiguous = retrieve_indices(out_reprs)

        self._weights_ranges = {}

        last_weight_position = 0

        self._contiguous = {}

        # iterate through the different groups of blocks
        # i.e., through all input/output pairs
        for io_pair in self._representations_pairs:

            self._contiguous[io_pair] = _in_contiguous[io_pair[0]] and _out_contiguous[io_pair[1]]

            # build the indices tensors
            if self._contiguous[io_pair]:
                in_indices = [
                    _in_indices[io_pair[0]].min(),
                    _in_indices[io_pair[0]].max() + 1,
                    _in_indices[io_pair[0]].max() + 1 - _in_indices[io_pair[0]].min()
                ]
                out_indices = [
                    _out_indices[io_pair[1]].min(),
                    _out_indices[io_pair[1]].max() + 1,
                    _out_indices[io_pair[1]].max() + 1 - _out_indices[io_pair[1]].min()
                ]

                setattr(self, 'in_indices_{}'.format(io_pair), in_indices)
                setattr(self, 'out_indices_{}'.format(io_pair), out_indices)

            else:
                out_indices, in_indices = _out_indices[io_pair[1]], _in_indices[io_pair[0]]
                # out_indices, in_indices = torch.meshgrid([_out_indices[io_pair[1]], _in_indices[io_pair[0]]])
                # in_indices = in_indices.reshape(-1)
                # out_indices = out_indices.reshape(-1)

                # register the indices tensors and the bases tensors as parameters of this module
                self.register_buffer('in_indices_{}'.format(io_pair), in_indices)
                self.register_buffer('out_indices_{}'.format(io_pair), out_indices)

            # number of occurrences of the input/output pair `io_pair`
            n_pairs = self._in_count[io_pair[0]] * self._out_count[io_pair[1]]

            # count the actual number of parameters
            total_weights = _block_sampler_modules[io_pair].dimension() * n_pairs

            # evaluate the indices in the global weights tensor to use for the basis belonging to this group
            self._weights_ranges[io_pair] = (last_weight_position, last_weight_position + total_weights)

            # increment the position counter
            last_weight_position += total_weights

        self._dim = last_weight_position

    def get_element_info(self, idx: int) -> Dict:

        assert 0 <= idx < self._dim, idx

        reprs_names = None
        relative_idx = None
        for pair, idx_range in self._weights_ranges.items():
            if idx_range[0] <= idx < idx_range[1]:
                reprs_names = pair
                relative_idx = idx - idx_range[0]
                break
        assert reprs_names is not None and relative_idx is not None

        block_sampler = getattr(self, f"block_sampler_{reprs_names}")
        block_idx = relative_idx // block_sampler.dimension()
        relative_idx = relative_idx % block_sampler.dimension()

        attr = block_sampler.get_element_info(relative_idx).copy()

        block_count = 0
        out_irreps_count = 0
        for o, o_repr in enumerate(self._out_reprs):
            in_irreps_count = 0
            for i, i_repr in enumerate(self._in_reprs):

                if reprs_names == (i_repr.name, o_repr.name):

                    if block_count == block_idx:
                        # retrieve the attributes of each basis element and build a new list of
                        # attributes adding information specific to the current block
                        attr.update({
                            "in_irreps_position": in_irreps_count + attr["in_irrep_idx"],
                            "out_irreps_position": out_irreps_count + attr["out_irrep_idx"],
                            "in_repr": reprs_names[0],
                            "out_repr": reprs_names[1],
                            "in_field_position": i,
                            "out_field_position": o,
                        })

                        attr['block_id'] = attr['id']
                        attr['id'] = idx

                        return attr

                    block_count += 1

                in_irreps_count += len(i_repr.irreps)
            out_irreps_count += len(o_repr.irreps)

        raise ValueError(f"Parameter with index {idx} not found!")

    def get_basis_info(self) -> Iterable[Dict]:

        out_irreps_counts = [0]
        out_block_counts = defaultdict(list)
        for o, o_repr in enumerate(self._out_reprs):
            out_irreps_counts.append(out_irreps_counts[-1] + len(o_repr.irreps))
            out_block_counts[o_repr.name].append(o)

        in_irreps_counts = [0]
        in_block_counts = defaultdict(list)
        for i, i_repr in enumerate(self._in_reprs):
            in_irreps_counts.append(in_irreps_counts[-1] + len(i_repr.irreps))
            in_block_counts[i_repr.name].append(i)

        # iterate through the different group of blocks
        # i.e., through all input/output pairs
        idx = 0
        for reprs_names in self._representations_pairs:

            block_sampler = getattr(self, f"block_sampler_{reprs_names}")

            # since this method returns an iterable of attributes built on the fly, load all attributes first and then
            # iterate on this list
            attrs = list(block_sampler.get_basis_info())

            for o in out_block_counts[reprs_names[1]]:
                out_irreps_count = out_irreps_counts[o]
                for i in in_block_counts[reprs_names[0]]:
                    in_irreps_count = in_irreps_counts[i]

                    # retrieve the attributes of each basis element and build a new list of
                    # attributes adding information specific to the current block
                    for attr in attrs:
                        attr = attr.copy()
                        attr.update({
                            "in_irreps_position": in_irreps_count + attr["in_irrep_idx"],
                            "out_irreps_position": out_irreps_count + attr["out_irrep_idx"],
                            "in_repr": reprs_names[0],
                            "out_repr": reprs_names[1],
                            "in_field_position": i,
                            "out_field_position": o,
                        })

                        # build the ids of the basis vectors
                        # add names and indices of the input and output fields
                        # id = '({}-{},{}-{})'.format(reprs_names[0], i, reprs_names[1], o)
                        # # add the original id in the block submodule
                        # id += "_" + attr["id"]
                        #
                        # # update with the new id
                        # attr["id"] = id
                        #
                        # attr["idx"] = idx

                        attr['block_id'] = attr['id']
                        attr['id'] = idx

                        assert idx < self._dim

                        idx += 1

                        yield attr

    def dimension(self) -> int:
        return self._dim

    def _compute_out_block(self, weights: torch.Tensor, input: torch.Tensor, points: torch.Tensor, io_pair) -> torch.Tensor:

        # retrieve the basis
        block_sampler = getattr(self, f"block_sampler_{io_pair}")

        # retrieve the linear coefficients for the basis sampler
        coefficients = weights[self._weights_ranges[io_pair][0]:self._weights_ranges[io_pair][1]]

        # reshape coefficients for the batch matrix multiplication
        coefficients = coefficients.view(
            self._out_count[io_pair[1]],  # u
            self._in_count[io_pair[0]],   # j
            block_sampler.dimension(),    # k
        )

        # expand the current subset of basis vectors and set the result in the appropriate place in the filter
        _basis = block_sampler(points)
        # p, o, i, k = _basis.shape

        # TODO: torch.einsum does not optimize the order of operations. Need to do this manually!
        _out = torch.einsum(
            'ujk,poik,pji->puo',
            coefficients,
            _basis,
            input,
        )

        return _out

    def _contract_basis_block(self, weights: torch.Tensor, points: torch.Tensor, io_pair) -> torch.Tensor:

        # retrieve the basis
        block_sampler = getattr(self, f"block_sampler_{io_pair}")

        # retrieve the linear coefficients for the basis sampler
        coefficients = weights[self._weights_ranges[io_pair][0]:self._weights_ranges[io_pair][1]]

        # reshape coefficients for the batch matrix multiplication
        coefficients = coefficients.view(
            self._out_count[io_pair[1]],  # u
            self._in_count[io_pair[0]],   # j
            block_sampler.dimension(),    # k
        )

        # expand the current subset of basis vectors and set the result in the appropriate place in the filter
        _basis = block_sampler(points)
        # p, o, i, k = _basis.shape

        _filter = torch.einsum(
            'ujk,poik->pujoi',
            coefficients,
            _basis,
        ).permute(0, 1, 3, 2, 4)

        return _filter

    def forward(self, weights: torch.Tensor, points: torch.Tensor) -> torch.Tensor:
        """
        Forward step of the Module which expands the basis, samples it on the input `points` and returns
        the filter built.

        Args:
            weights (torch.Tensor): the learnable weights used to linearly combine the basis filters
            points (torch.Tensor): the points where the filter should be sampled

        Returns:
            the filter built

        """
        assert weights.shape[0] == self.dimension()
        assert len(weights.shape) == 1

        S = points.shape[0]

        if self._uniform:
            # if there is only one block (i.e. one type of input field and one type of output field),
            # we can return the computed block immediately, instead of copying it inside a preallocated empty tensor
            io_pair = self._representations_pairs[0]
            _filter = self._contract_basis_block(weights, points, io_pair)
            _filter = _filter.reshape(S, self._output_size, self._input_size)

        else:

            # to support Automatic Mixed Precision (AMP), we can not preallocate the output tensor with a specific dtype
            # Instead, we check the dtype of the first `expanded` block. For this reason, we postpose the allocation
            # of the full _filter tensor

            _filter = None

            # iterate through all input-output field representations pairs
            for io_pair in self._representations_pairs:

                # retrieve the indices
                in_indices = getattr(self, f"in_indices_{io_pair}")
                out_indices = getattr(self, f"out_indices_{io_pair}")

                # expand the current subset of basis vectors and set the result in the appropriate place in the filter
                expanded = self._contract_basis_block(weights, points, io_pair)

                if _filter is None:
                    # build the tensor which will contain the filter
                    # this lazy strategy allows us to use expanded.dtype which is dynamically chosen by PyTorch's AMP
                    _filter = torch.zeros(
                        S, self._output_size, self._input_size,
                        device=weights.device,
                        dtype=expanded.dtype,
                    )

                if self._contiguous[io_pair]:
                    _filter[
                        :,
                        out_indices[0]:out_indices[1],
                        in_indices[0]:in_indices[1],
                    ] = expanded.reshape(S, out_indices[2], in_indices[2])
                else:
                    out_indices, in_indices = torch.meshgrid([out_indices, in_indices])
                    in_indices = in_indices.reshape(-1)
                    out_indices = out_indices.reshape(-1)
                    _filter[
                        :,
                        out_indices,
                        in_indices,
                    ] = expanded.reshape(S, -1)

            if _filter is None:
                # just in case
                _filter = torch.zeros(
                    S, self._output_size, self._input_size,
                    device=weights.device,
                    dtype=weights.dtype,
                )

        # return the new filter
        return _filter

    def _expand_filter_then_compute(self, weights: torch.Tensor, input: torch.Tensor, points: torch.Tensor) -> torch.Tensor:

        filter = self(weights, points)

        return torch.einsum(
            'poi,pi->po',
            filter, input
        )

    def _compute_then_expand_filter(self, weights: torch.Tensor, input: torch.Tensor,
                                    points: torch.Tensor) -> torch.Tensor:

        S = input.shape[0]
        assert S > 0, S

        if self._uniform:

            # if there is only one block (i.e. one type of input field and one type of output field),
            #  we can return the computed block immediately, instead of copying it inside a preallocated empty tensor
            io_pair = self._representations_pairs[0]
            in_repr = io_pair[0]
            _input = input.view(S, self._in_count[in_repr], self._in_sizes[in_repr])
            return self._compute_out_block(weights, _input, points, io_pair).reshape(S, self._output_size)

        else:

            # build the tensor which will contain the output
            _out = torch.zeros(S, self._output_size, device=input.device, dtype=input.dtype)

            # iterate through all input-output field representations pairs
            for io_pair in self._representations_pairs:

                # retrieve the indices
                in_indices = getattr(self, f"in_indices_{io_pair}")
                out_indices = getattr(self, f"out_indices_{io_pair}")

                if self._contiguous[io_pair]:
                    _input = input[:, in_indices[0]:in_indices[1]]
                else:
                    _input = input[:, in_indices]

                in_repr = io_pair[0]
                _input = _input.view(S, self._in_count[in_repr], self._in_sizes[in_repr])
                # expand the current subset of basis vectors and set the result in the appropriate place in the filter
                block_out = self._compute_out_block(weights, _input, points, io_pair)

                if self._contiguous[io_pair]:
                    _out[:, out_indices[0]:out_indices[1]] += block_out.reshape(S, out_indices[2])
                else:
                    _out[:, out_indices] += block_out.reshape(S, -1)

            return _out

    def compute_messages(self,
                         weights: torch.Tensor,
                         input: torch.Tensor,
                         points: torch.Tensor,
                         conv_first: bool = True,
                ) -> torch.Tensor:
        """
        Expands the basis with the learnable weights to generate the filter and use it to compute the messages along the
        edges.

        Each point in `points` corresponds to an edge in a graph.
        Each point is associated with a row of `input`.
        This row is a feature associated to the source node of the edge which needs to be propagated to the target
        node of the edge.

        Args:
            weights (torch.Tensor): the learnable weights used to linearly combine the basis filters
            input (torch.Tensor): the input features associated with each point
            points (torch.Tensor): the points where the filter should be sampled
            conv_first (bool, optional): perform convolution with the basis filters and, then, combine the responses
                                         with the learnable weights. This generally has computational benefits.
                                         (Default ``True``).

        Returns:
            the messages computed

        """
        assert weights.shape[0] == self.dimension()
        assert len(weights.shape) == 1

        assert len(input.shape) == 2
        assert input.shape[1] == self._input_size
        assert input.shape[0] == points.shape[0]

        if conv_first:
            return self._compute_then_expand_filter(weights, input, points)
        else:
            return self._expand_filter_then_compute(weights, input, points)

    def __hash__(self):

        _hash = 0
        for io in self._representations_pairs:
            n_pairs = self._in_count[io[0]] * self._out_count[io[1]]
            _hash += hash(getattr(self, f"block_sampler_{io}")) * n_pairs

        return _hash

    def __eq__(self, other):
        if not isinstance(other, BlocksBasisSampler):
            return False

        if self._dim != other._dim:
            return False

        if self._representations_pairs != other._representations_pairs:
            return False

        for io in self._representations_pairs:
            if self._contiguous[io] != other._contiguous[io]:
                return False

            if self._weights_ranges[io] != other._weights_ranges[io]:
                return False

            if self._contiguous[io]:
                if getattr(self, f"in_indices_{io}") != getattr(other, f"in_indices_{io}"):
                    return False
                if getattr(self, f"out_indices_{io}") != getattr(other, f"out_indices_{io}"):
                    return False
            else:
                if torch.any(getattr(self, f"in_indices_{io}") != getattr(other, f"in_indices_{io}")):
                    return False
                if torch.any(getattr(self, f"out_indices_{io}") != getattr(other, f"out_indices_{io}")):
                    return False

            if getattr(self, f"block_sampler_{io}") != getattr(other, f"block_sampler_{io}"):
                return False

        return True
